__auth__ = 'A.Star'

import numpy as np
from qgis.core import QgsDistanceArea, QgsUnitTypes, QgsPointXY
from snowland.gis_tool import ELLIPSOID

npa = np.array


class DistanceArea:
    def __init__(self, ellipsoid=ELLIPSOID, x=None, y=None):
        self.__qgs_distance = QgsDistanceArea()
        if x is None or y is None:
            if isinstance(ellipsoid, (list, tuple, np.ndarray)):
                self.__qgs_distance.setEllipsoid(*ellipsoid[:2])
            else:
                self.__qgs_distance.setEllipsoid(*ellipsoid[:2])
        else:
            self.__qgs_distance.setEllipsoid(x, y)

    def set_ellipsoid(self, x, y):
        self.__qgs_distance.setEllipsoid(x, y)

    def distance_centimetres(self, Olon, Olat, Dlon, Dlat):
        """
        通过两经纬度点计算地表每一段实际距离, 单位是厘米
        """
        if isinstance(Olon, (float, np.float64)):
            p1, p2 = QgsPointXY(Olon, Olat), QgsPointXY(Dlon, Dlat)
            return self.__qgs_distance.convertLengthMeasurement(self.__qgs_distance.measureLine(p1, p2),
                                                                QgsUnitTypes.DistanceCentimeters)
        else:
            args = [(QgsPointXY(*p1[:2]), QgsPointXY(*p2[:2])) for p1, p2 in zip(zip(Olon, Olat), zip(Dlon, Dlat))]
            return npa([self.__qgs_distance.convertLengthMeasurement(self.__qgs_distance.measureLine(p1, p2),
                                                                     QgsUnitTypes.DistanceCentimeters)
                        for p1, p2 in args])

    def distance_metres(self, Olon, Olat, Dlon, Dlat):
        """
        通过两经纬度点计算地表每一段实际距离, 单位是米
        """
        if isinstance(Olon, (float, np.float64)):
            p1, p2 = QgsPointXY(Olon, Olat), QgsPointXY(Dlon, Dlat)
            return self.__qgs_distance.convertLengthMeasurement(self.__qgs_distance.measureLine(p1, p2),
                                                                QgsUnitTypes.DistanceMeters)
        else:
            args = [(QgsPointXY(*p1[:2]), QgsPointXY(*p2[:2])) for p1, p2 in zip(zip(Olon, Olat), zip(Dlon, Dlat))]
            return npa([self.__qgs_distance.convertLengthMeasurement(self.__qgs_distance.measureLine(p1, p2),
                                                                     QgsUnitTypes.DistanceMeters)
                        for p1, p2 in args])

    def length_centimetres(self, Olon, Olat):
        """
        通过两经纬度点计算地表实际距离之和
        """
        return np.sum(self.distance_centimetres(Olon[:-1], Olat[:-1], Olon[1:], Olat[1:]))

    def length_metres(self, Olon, Olat):
        """
        通过两经纬度点计算地表实际距离
        """
        return np.sum(self.distance_metres(Olon[:-1], Olat[:-1], Olon[1:], Olat[1:]))

    def distance_centimetres_by_points(self, ps1, ps2):
        """
        通过两经纬度点计算地表实际距离，单位是厘米
        """
        args = [(QgsPointXY(*p1[:2]), QgsPointXY(*p2[:2])) for p1, p2 in zip(ps1, ps2)]
        return npa([self.__qgs_distance.convertLengthMeasurement(self.__qgs_distance.measureLine(p1, p2),
                                                                 QgsUnitTypes.DistanceCentimeters)
                    for p1, p2 in args])

    def distance_metres_by_points(self, ps1, ps2):
        """
        通过两经纬度点计算地表实际距离,单位是米
        """
        args = [(QgsPointXY(*p1[:2]), QgsPointXY(*p2[:2])) for p1, p2 in zip(ps1, ps2)]
        return npa([self.__qgs_distance.convertLengthMeasurement(self.__qgs_distance.measureLine(p1, p2),
                                                                 QgsUnitTypes.DistanceMeters)
                    for p1, p2 in args])

    def distance_centimetres_by_two_points(self, p1, p2):
        """
        通过两经纬度点计算地表实际距离， 单位是厘米
        """
        if isinstance(p1, np.ndarray):
            p1 = QgsPointXY(*p1[:2])
        if isinstance(p2, np.ndarray):
            p2 = QgsPointXY(*p2[:2])
        return self.__qgs_distance.convertLengthMeasurement(self.__qgs_distance.measureLine(p1, p2),
                                                            QgsUnitTypes.DistanceCentimeters)

    def distance_metres_by_two_points(self, p1, p2):
        """通过两经纬度点计算地表实际距离， 单位是厘米"""
        if isinstance(p1, np.ndarray):
            p1 = QgsPointXY(*p1[:2])
        if isinstance(p2, np.ndarray):
            p2 = QgsPointXY(*p2[:2])
        return self.__qgs_distance.convertLengthMeasurement(self.__qgs_distance.measureLine(p1, p2),
                                                            QgsUnitTypes.DistanceMeters)





EARTH_RADIUS = 6378.137  # 初始化地球半径km
ELLIPSOID = (6378137, 6356752.314245179)

RULE_MORTON = (1 << 32) / 360.0
RULE_MORTON_TO_LONLAT = 360.0 / (1 << 32)
NDS_90_DEGREES = 0x3fffffff
NDS_180_DEGREES = 0x7fffffff
NDS_360_DEGREES = 4294967295

from snowland.gis_tool import gis_helper, haversine, nds, transformer



from qgis.core import QgsGeometry

from snowland.gis_tool.qgis_tool.qgis_util import vector, vector_end, vector_start,\
    points, pointsXY, pointsZM, start_pointXY, end_pointXY
from snowland.gis_tool.qgis_tool import distance_area, qgis_util, utils

QgsGeometry.points = points
QgsGeometry.pointsXY = pointsXY
QgsGeometry.pointsZM = pointsZM
QgsGeometry.vector = vector
QgsGeometry.vector_start = vector_start
QgsGeometry.vector_end = vector_end
QgsGeometry.start_pointXY = start_pointXY
QgsGeometry.end_pointXY = end_pointXY



import numpy as np
from scipy.spatial.distance import euclidean

npa = np.array


class GisHelper:
    @staticmethod
    def get_point_by_rate(line: np.ndarray, meters, metric=euclidean):
        """
        在line组成的折线中,获得距离起点距离为metres的点
        """
        return GisHelper.get_point_by_rate_index(line, meters, metric)[0]

    @staticmethod
    def get_point_by_rate_index(line: np.ndarray, meters, metric=euclidean):
        """
        在line组成的折线中,获得距离起点距离为metres的点和小于这个点的最大节点编号
        """
        meters_all = npa([metric(a, b) for a, b in zip(line[:-1], line[1:])])
        s = np.cumsum(meters_all)
        if s[-1] < meters:
            return line[-1], len(s)
        ind = np.sum(s < meters_all)
        s = np.insert(s, 0, 0)
        delta_meters = meters - s[ind]
        vector_line = line[1:] - line[:-1]
        return line[ind] + delta_meters / meters_all[ind] * vector_line[ind], ind


def get_point_by_rate_index(line: np.ndarray, meters, metric=euclidean):
    """
    在line组成的折线中,获得距离起点距离为metres的点和小于这个点的最大节点编号
    :param
    """
    meters_all = npa([metric(a, b) for a, b in zip(line[:-1], line[1:])])
    s = np.cumsum(meters_all)
    if s[-1] < meters:
        return line[-1], len(s)
    ind = np.sum(s < meters_all)
    s = np.insert(s, 0, 0)
    delta_meters = meters - s[ind]
    vector_line = line[1:] - line[:-1]
    return line[ind] + delta_meters / meters_all[ind] * vector_line[ind], ind


def get_point_by_rate(line: np.ndarray, meters, metric=euclidean):
    """
    在line组成的折线中,获得距离起点距离为metres的点
    """
    return get_point_by_rate_index(line, meters, metric)[0]



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: utils.py
# @time: 2022/01/06 20:48
# @Software: PyCharm
from typing import List

import numpy as np
from scipy.spatial.distance import cdist, pdist, euclidean
from astartool.number import equals_zero, equals_zero_all
from astartool.data_structure import LinkedList

npa = np.array
npl = np.linalg


def middle(x, x1, x2, eps=1e-8):
    """
    判断点是否在x1, x2之间
    x, x1, x2 均为ndarray
    """
    dx1 = x - x1
    dx2 = x - x2
    return ((x1 < x) & (x < x2)) | ((x1 > x) & (x > x2)) | ((-eps < dx1) & (dx1 < eps)) | ((-eps < dx2) & (dx2 < eps))


def is_in_between(x, x1, x2, eps=1e-8):
    """
    判断点是否在x1, x2之间
    x, x1, x2 均为ndarray
    """
    dx1 = x - x1
    dx2 = x - x2
    return ((x1 < x) & (x < x2)) | ((x1 > x) & (x > x2)) | ((-eps < dx1) & (dx1 < eps)) | ((-eps < dx2) & (dx2 < eps))


def get_lines(line_points):
    """
    """
    A = line_points[1:, 1] - line_points[:-1, 1]
    B = line_points[:-1, 0] - line_points[1:, 0]
    C = line_points[1:, 0] * line_points[:-1, 1] - line_points[:-1, 0] * line_points[1:, 1]
    return np.vstack((A, B, C)).T


def get_foot(point, line=None, A=None, B=None, C=None):
    """
    point: 单点, ndarray/list len(points) = 2
    line: Ax+By+C=0
    返回 foot_x, foot_y, 垂足坐标

    """
    if line is not None:
        A, B, C = line[:, 0], line[:, 1], line[:, 2]
    elif A is not None and B is not None and C is not None:
        pass

    x, y = point[0], point[1]
    A2, B2, AB, AC, BC = A * A, B * B, A * B, A * C, B * C
    # foot_x, foot_y 即为垂足坐标
    foot_x = (B2 * x - AB * y - AC) / (A2 + B2)
    foot_y = (A2 * y - AB * x - BC) / (A2 + B2)

    return foot_x, foot_y


def get_intersect_point(line):
    """
    line_points 是 n x 3 矩阵，代表组成线标准式的参数 Ax+By+C=0 的系数 ABC
    返回这些线的交点
    """
    line1 = line[:-1, :]
    line2 = line[1:, :]

    a1 = line1[:, 0]
    b1 = line1[:, 1]
    c1 = line1[:, 2]
    a2 = line2[:, 0]
    b2 = line2[:, 1]
    c2 = line2[:, 2]
    b1c2 = b1 * c2
    b2c1 = b2 * c1
    a1c2 = a1 * c2
    a1b2 = a1 * b2
    a2b1 = a2 * b1
    a2c1 = a2 * c1

    x = (b1c2 - b2c1) / (a1b2 - a2b1)
    y = (a2c1 - a1c2) / (a1b2 - a2b1)
    return x, y


def distance_point_line(points, line_points, metric=euclidean, eps=1e-8):
    """
    points: 单点, ndarray/list len(points) = 2
    line_points: 每两点组成的一条线， 代表一条linestring
    返回tuple(isVe, ind, d)
    isVe: 找到的最近点是不是线的顶点
    p: 距离最近的点，ndarray
    d: 最近距离是多少
    """
    # ABC即为线的参数 Ax + By + C = 0
    A = line_points[1:, 1] - line_points[:-1, 1]
    B = line_points[:-1, 0] - line_points[1:, 0]
    C = line_points[1:, 0] * line_points[:-1, 1] - line_points[:-1, 0] * line_points[1:, 1]

    foot_x, foot_y = get_foot(points, A=A, B=B, C=C)
    # 是否为顶点坐标
    is_vertex = True

    index = is_in_between(foot_x, line_points[1:, 0], line_points[:-1, 0], eps=eps) & \
            is_in_between(foot_y, line_points[1:, 1], line_points[:-1, 1], eps=eps)
    if np.any(index):
        is_vertex = False
        points_check = np.vstack((foot_x[index].T, foot_y[index].T)).T
    else:
        points_check = line_points

    dist = cdist(npa([points]), points_check, metric=metric)
    ind = np.argmin(dist)
    return is_vertex, points_check[ind], dist[0, ind]


def distance_point_line_index(point, line_points, metric=euclidean, eps=1e-8):
    """
    params points: 单点(x, y), ndarray/list len(points) = 2
    line_points: 每两点组成的一条线， 代表一条linestring
    返回tuple(isVe, ind, d)
    is_vertex: 找到的最近点是不是线的顶点
    p: 距离最近的点，ndarray
    d: 最近距离是多少
    ind: index 若 is_vertex 为 True, ind为最近节点的编号, 若 is_vertex 为 False, ind为最近节点的编号

    """
    # ABC即为线的参数 Ax + By + C = 0
    A = line_points[1:, 1] - line_points[:-1, 1]
    B = line_points[:-1, 0] - line_points[1:, 0]
    C = line_points[1:, 0] * line_points[:-1, 1] - line_points[:-1, 0] * line_points[1:, 1]

    foot_x, foot_y = get_foot(point, A=A, B=B, C=C)
    # 是否为顶点坐标
    is_vertex = True

    index = middle(foot_x, line_points[1:, 0], line_points[:-1, 0], eps=eps) & \
            middle(foot_y, line_points[1:, 1], line_points[:-1, 1], eps=eps)
    if np.any(index):
        is_vertex = False
        index_ind, = np.where(index)
        points_check = np.vstack((foot_x[index].T, foot_y[index].T)).T

        dist = cdist(npa([point]), points_check, metric=metric)
        ind = np.argmin(dist)
        return is_vertex, points_check[ind], dist[0, ind], index_ind[ind]
    else:
        points_check = line_points
        dist = cdist(npa([point]), points_check, metric=metric)
        ind = np.argmin(dist)
        return is_vertex, points_check[ind], dist[0, ind], ind


def move_distance(a, dist, flag):
    """
    移线
    """
    lines = get_lines(a)
    norm_line = npl.norm(lines[:, :2], axis=1)
    delta_c = dist * norm_line
    lines_new = lines
    if flag:
        lines_new[:, 2] -= delta_c
    else:
        lines_new[:, 2] += delta_c
    x, y = get_intersect_point(lines_new)
    if len(x):
        x_start, y_start = get_foot(a[0, :], lines[:1, :])
        x_end, y_end = get_foot(a[-1, :], lines[-1:, :])
        x = np.hstack((x_start, x, x_end))
        y = np.hstack((y_start, y, y_end))
    return x, y


def simple_line(points=None, x: np.ndarray = None, y: np.ndarray = None, forward_point=8):
    """
    对一条线(x, y)进行化简，删除折返线
    """
    if x is not None and y is not None:
        x_array = npa(x)
        y_array = npa(y)
        points = np.vstack((x_array, y_array)).T

    length_points = len(points)
    ps_index = np.empty(length_points, dtype=int)
    i, cnt = 0, 0
    ps = [points[0]]
    while i + 1 < length_points:
        p1 = points[i]
        p2 = points[i + 1]
        p_temp = p2
        for j in range(min(i + forward_point, length_points - 1), i + 1, -1):
            p3 = points[j]
            p4 = points[j - 1]
            x, y = get_intersect_by_two_point(p1, p2, p3, p4)
            if x is not None and y is not None:
                p_temp = x, y
                global_j = j - 1
                break
        else:
            global_j = i + 1
        ps_index[i] = cnt
        ps_index[i + 1:global_j] = i + 1
        cnt += 1
        i = global_j
        ps.append(p_temp)
    ps_index[-1] = cnt
    ps_array = npa(ps)
    return ps_array[:, 0], ps_array[:, 1], ps_index


def move_distance_with_endpoints(a, dist, flag):
    """
    移线，注意和move_distance移线不同！！！
    此函数带有端点
    """
    lines = get_lines(a)
    norm_line = npl.norm(lines[:, :2], axis=1)
    delta_c = dist * norm_line
    lines_new = lines
    if flag:
        lines_new[:, 2] -= delta_c
    else:
        lines_new[:, 2] += delta_c
    x, y = get_intersect_point(lines_new)
    x_start, y_start = get_foot(a[0, :], lines[:1, :])
    x_end, y_end = get_foot(a[-1, :], lines[-1:, :])
    x = np.hstack((x_start, x, x_end))
    y = np.hstack((y_start, y, y_end))
    return x, y


def interp_points(line1: np.ndarray, line2: np.ndarray, metric=euclidean, eps=1e-8):
    """
    插入点
    :param line1:
    :param line2:
    :param metric:
    :param eps:
    :return:
    """

    line1 = line1[:, :2]
    line2 = line2[:, :2]

    line1, line2 = alignment(line1, line2, metric, eps=eps)

    point_list_2 = LinkedList()
    new_point_list_2 = []
    for p in line1[1:-1]:
        flag, point, dist, index = distance_point_line_index(p, line2, metric=metric, eps=eps)
        if not flag:
            point_list_2.append((index, point))

    point_list_1 = LinkedList()
    new_point_list_1 = []
    for p in line2[1:-1]:
        flag, point, dist, index = distance_point_line_index(p, line1, metric=metric, eps=eps)
        if not flag:
            point_list_1.append((index, point))

    pre_i = 0
    for item in point_list_1:
        i, p = item
        new_point_list_1.extend(line1[pre_i:i + 1])
        new_point_list_1.append(p)
        pre_i = i + 1
    new_point_list_1.extend(line1[pre_i:])

    pre_j = 0
    for item in point_list_2:
        j, p = item
        new_point_list_2.extend(line2[pre_j:j + 1])
        new_point_list_2.append(p)
        pre_j = j + 1
    new_point_list_2.extend(line2[pre_j:])

    # assert len(new_point_list_1) == len(new_point_list_2)
    return np.vstack(new_point_list_1), np.vstack(new_point_list_2)


def alignment(linestring1: np.ndarray, linestring2: np.ndarray, metric=euclidean, eps: float = 1.0):
    """
    化简车道线,使之按垂直方向对齐
    linestring1: 第一根线 m x 2 ndarray
    linestring2: 第二根线 n x 2 ndarray
    metric: 距离计算函数
    eps: 若距离小于 eps, 则不需要截取
    """
    assert linestring1.shape[1] == 2
    assert linestring2.shape[1] == 2
    is_vertex, points_check, dist, ind = distance_point_line_index(linestring2[0], linestring1, metric)
    if (not is_vertex) and metric(points_check, linestring1[0]) > eps:
        linestring1 = linestring1[ind + 1:]
        linestring1 = np.vstack((points_check, linestring1))
    else:
        is_vertex, points_check, dist, ind = distance_point_line_index(linestring1[0], linestring2, metric)
        if (not is_vertex) and metric(points_check, linestring2[0]) > eps:
            linestring2 = linestring2[ind + 1:]
            linestring2 = np.vstack((points_check, linestring2))

    is_vertex, points_check, dist, ind = distance_point_line_index(linestring2[-1], linestring1, metric)
    if (not is_vertex) and metric(points_check, linestring1[-1]) > eps:
        linestring1 = linestring1[:ind + 1]
        linestring1 = np.vstack((linestring1, points_check))
    else:
        is_vertex, points_check, dist, ind = distance_point_line_index(linestring1[-1], linestring2, metric)
        if (not is_vertex) and metric(points_check, linestring2[-1]) > eps:
            linestring2 = linestring2[:ind + 1]
            linestring2 = np.vstack((linestring2, points_check))

    return linestring1, linestring2


def rotate_geometry(poly: (np.ndarray, List), rad):
    """
    二维旋转rad弧度
    """
    if isinstance(poly, List):
        poly = npa(poly)
    if len(poly.shape) == 1:
        flag = True
        poly = np.expand_dims(poly, 0)
    else:
        flag = False
    matrix = npa([[np.cos(rad), np.sin(rad)],
                  [-np.sin(rad), np.cos(rad)]])
    new_poly = np.dot(poly, matrix)
    return new_poly[0, :] if flag else new_poly


def bounding_box(poly: np.ndarray, eps=1e-10):
    """
    boundingBox
    """
    # shape of min_rect: [4, 2]
    x_min, y_min = np.min(poly, axis=0)
    x_max, y_max = np.max(poly, axis=0)
    return npa([x_min, y_min]), (x_max - x_min), (y_max - y_min)


def min_rotate_rect_a(hull: np.ndarray, eps=1e-10):
    """
    最小外接矩形--按面积
    """
    area = np.inf
    lines_com = []
    if not equals_zero_all(hull[0] - hull[-1], eps):
        points_hull = np.vstack((hull, hull[0]))
    else:
        points_hull = hull

    lines = get_lines(points_hull)
    for i, p2 in enumerate(points_hull[:-1]):
        line = lines[i]
        dist_cmp = (points_hull[:, 0] * line[0] + points_hull[:, 1] * line[1] + line[2])
        dist = dist_cmp / npl.norm(line[:2])

        h_ind = np.argmax(np.abs(dist))  # 得到距离最大的点距离，即为高，同时得到该点坐标
        h = dist[h_ind]
        line_px = line.copy()
        line_px[2] -= dist_cmp[h_ind]
        line_cz = npa([line[1], -line[0], 0])

        dist_cmp = (points_hull[:, 0] * line_cz[0] + points_hull[:, 1] * line_cz[1])
        dist = dist_cmp / npl.norm(line_cz[:2])

        v_ind_max = np.argmax(dist)  # 得到距离最大的点距离，即为高
        v_ind_min = np.argmin(dist)

        w = dist[v_ind_max] - dist[v_ind_min]
        a = abs(h * w)
        if area >= h * w:
            # 使面积最小
            area = a
            line_cz_1 = line_cz.copy()
            line_cz_1[2] -= dist_cmp[v_ind_max]
            line_cz_2 = line_cz.copy()
            line_cz_2[2] = dist_cmp[v_ind_min]
            lines_com = [line, line_cz_1, line_px, line_cz_2, line]

    x, y = get_intersect_point(np.vstack(lines_com))
    return np.vstack((x, y)).T, area


def min_rotate_rect_c(hull: np.ndarray, eps=1e-10):
    """
    最小外接矩形--按周长
    """
    cir = np.inf
    lines_com = []
    if not equals_zero_all(hull[0] - hull[-1], eps):
        points_hull = np.vstack((hull, hull[0]))
    else:
        points_hull = hull

    lines = get_lines(points_hull)
    for i, p2 in enumerate(points_hull[:-1]):
        line = lines[i]
        dist_cmp = (points_hull[:, 0] * line[0] + points_hull[:, 1] * line[1] + line[2])
        dist = dist_cmp / npl.norm(line[:2])

        h_ind = np.argmax(np.abs(dist))  # 得到距离最大的点距离，即为高，同时得到该点坐标
        h = dist[h_ind]
        line_px = line.copy()
        line_px[2] -= dist_cmp[h_ind]
        line_cz = npa([line[1], -line[0], 0])

        dist_cmp = (points_hull[:, 0] * line_cz[0] + points_hull[:, 1] * line_cz[1])
        dist = dist_cmp / npl.norm(line_cz[:2])

        v_ind_max = np.argmax(dist)  # 得到距离最大的点距离，即为高
        v_ind_min = np.argmin(dist)

        w = abs(dist[v_ind_max] - dist[v_ind_min])
        c = h + w
        if cir >= c:
            # 使面积最小
            cir = c
            line_cz_1 = line_cz.copy()
            line_cz_1[2] -= dist_cmp[v_ind_max]
            line_cz_2 = line_cz.copy()
            line_cz_2[2] = dist_cmp[v_ind_min]
            lines_com = [line, line_cz_1, line_px, line_cz_2, line]

    x, y = get_intersect_point(np.vstack(lines_com))
    return np.vstack((x, y)).T, cir * 2


def min_rotate_rect(hull: np.ndarray, cmp: str = 'a', eps=1e-10):
    """
    最小外接矩形
    """
    if cmp.lower().startswith('a'):
        return min_rotate_rect_a(hull, eps)
    else:
        return min_rotate_rect_c(hull, eps)


def get_angle_rad(a, b, eps=1e-20):
    """
    a, b 为一维向量， 多点会错！！！
    返回 向量a和b之间的夹角， 值域是 -pi ~ pi
    """
    a = npa(a)
    b = npa(b)
    norm_a = npl.norm(a, axis=1) if len(a.shape) == 2 else npl.norm(a)
    norm_b = npl.norm(b, axis=1) if len(b.shape) == 2 else npl.norm(b)
    # 单位化（可以不用这一步）
    # a = a / norm_a  # 不能写成 a /= norm_a
    # b = b / norm_b  # 不能写成 b /= norm_b
    # 夹角cos值
    if -eps < norm_a < eps or -eps < norm_b < eps:
        return np.nan
    cos_ = np.dot(a, b) / (norm_a * norm_b)
    # 夹角sin值
    sin_ = np.cross(a, b) / (norm_a * norm_b)
    arctan2_ = np.arctan2(sin_, cos_)
    return arctan2_


def get_intersect_by_two_point(p1, p2, p3, p4):
    """
    判断线段是否相交
    """
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    x4, y4 = p4
    b1 = (y2 - y1) * x1 + (x1 - x2) * y1
    b2 = (y4 - y3) * x3 + (x3 - x4) * y3
    D = (x2 - x1) * (y4 - y3) - (x4 - x3) * (y2 - y1)
    D1 = b2 * (x2 - x1) - b1 * (x4 - x3)
    D2 = b2 * (y2 - y1) - b1 * (y4 - y3)
    x, y = D1 / D, D2 / D
    if middle(x, x1, x2) and middle(x, x3, x4) and middle(y, y1, y2) and middle(y, y3, y4):
        return x, y
    else:
        return None, None


def get_angle_degree(a: (list, np.ndarray), b: (list, np.ndarray)):
    """
    返回量向量之间的夹角，值域是 0~180，单位是度
    """
    # 初始化向量
    degree = np.rad2deg(get_angle_rad(a, b))
    return np.abs(degree)


def get_rotate_angle_degree(v1, v2):
    """
    v1旋转到v2经历的角度， 值域0~360, 单位是度
    """
    return np.rad2deg(get_angle_rad(v1, v2)) % 360


def get_rotate_angle_rad(v1, v2):
    """
    v1旋转到v2经历的角度， 值域0~2*pi, 单位是弧度
    """
    return get_angle_rad(v1, v2) % (2 * np.pi)


def get_point_by_rate(line: np.ndarray, meters, metric=euclidean):
    """
    在line组成的折线中,获得距离起点距离为metres的点
    """
    return get_point_by_rate_index(line, meters, metric)[0]


def get_point_by_rate_index(line: np.ndarray, meters, metric=euclidean):
    """
    在line组成的折线中,获得距离起点距离为metres的点和小于这个点的最大节点编号
    """
    meters_all = npa([metric(a, b) for a, b in zip(line[:-1], line[1:])])
    s = np.cumsum(meters_all)
    if s[-1] < meters:
        return line[-1], len(s)
    ind = np.sum(s < meters_all)
    s = np.insert(s, 0, 0)
    delta_meters = meters - s[ind]
    vector_line = line[1:] - line[:-1]
    return line[ind] + delta_meters / meters_all[ind] * vector_line[ind], ind


def get_arc(o, r, theta1, theta2, length):
    """
    """
    theta = np.linspace(theta1, theta2, length)
    x = r * np.cos(theta) + o[0]
    y = r * np.sin(theta) + o[1]
    return x, y


def rect_to_points(rect):
    """
    rect矩形转坐标点
    """
    points = np.vstack([rect[0],
                        [rect[0][0], rect[0][1] + rect[2]],
                        [rect[0][0] + rect[1], rect[0][1] + rect[2]],
                        [rect[0][0] + rect[1], rect[0][1]],
                        rect[0]
                        ])
    return points


def move_distance_by_point(p, vector, dist, metric=euclidean):
    unit_vector = vector / npl.norm(vector)
    unit_point = unit_vector + p
    unitLen = metric(p, unit_point)
    vecCenter = dist * unit_vector / unitLen
    return vecCenter + p


def move_distance_for_polygon(points, dist, flag=1, metric=euclidean):
    sign = np.sign(flag)
    lines = []
    for p1, p2 in zip(points[:-1, :], points[1:, :]):
        v_edge = p2 - p1
        if equals_zero_all(v_edge):
            continue
        v = rotate_geometry(v_edge, -sign * np.pi / 2)
        new_p1 = move_distance_by_point(p1, v, dist, metric)
        new_p2 = move_distance_by_point(p2, v, dist, metric)
        lines.append(get_lines(np.vstack([new_p1, new_p2])))
    lines.append(lines[0])
    lines_ndarray = np.vstack(lines)
    x, y = get_intersect_point(lines_ndarray)
    return np.hstack((x[-1], x)), np.hstack((y[-1], y))


def curvature(ps):
    """
    计算曲率
    :param ps n x 2 ndarray
    :return:
    """
    vs = ps[1:] - ps[:-1]
    vs2 = ps[2:] - ps[:-2]
    dis1 = np.linalg.norm(vs, axis=1)
    dis2 = np.linalg.norm(vs2, axis=1)
    return 2 * (vs[:-1, 0] * vs[1:, 1] - vs[:-1, 1] * vs[1:, 0]) / (dis1[:-1] * dis1[1:] * dis2)



# -*- coding: utf-8 -*-

import numpy as np
from pyproj import Transformer
from typing import Iterable


npa = np.array


def utm_transformer(latitude, longitude):
    """
    4326转utm坐标系
    """
    utm_x = np.empty_like(latitude)
    utm_y = np.empty_like(longitude)
    latitude = npa(latitude)
    longitude = npa(longitude)
    zone_number = (np.array(longitude) / 6).astype(int) + 31
    if isinstance(zone_number, Iterable):
        zone_set = set(zone_number)
        for zone_num in zone_set:
            utm_source_str = f"EPSG:326{zone_num:02}"
            ind = zone_number == zone_num
            transformer = Transformer.from_crs("EPSG:4326", utm_source_str)
            utm_x[ind], utm_y[ind] = transformer.transform(latitude[ind], longitude[ind])
    else:
        utm_source_str = f"EPSG:326{zone_number:02}"
        transformer = Transformer.from_crs("EPSG:4326", utm_source_str)
        utm_x, utm_y = transformer.transform(latitude, longitude)
    return utm_x, utm_y


def latlon_transformer(utm_x, utm_y, zone_number):
    """
    utm转4326坐标系
    """
    latitude = np.empty_like(utm_x)
    longitude = np.empty_like(utm_y)
    utm_x = npa(utm_x)
    utm_y = npa(utm_x)

    if isinstance(zone_number, Iterable):
        zone_set = set(zone_number)
        for zone_num in zone_set:
            utm_source_str = f"EPSG:326{zone_num:02}"
            ind = zone_number == zone_num
            transformer = Transformer.from_crs(utm_source_str, "EPSG:4326")
            latitude[ind], longitude[ind] = transformer.transform(utm_x[ind], utm_y[ind])
    else:
        utm_source_str = f"EPSG:326{zone_number:02}"
        transformer = Transformer.from_crs(utm_source_str, "EPSG:4326")
        latitude, longitude= transformer.transform(utm_x, utm_y)
    return latitude, longitude



import configparser
import pymysql


cp = configparser.ConfigParser()


class MysqlConnection:

    def __init__(self, session, file_name='db.conf'):
        cp.read(file_name)
        self.database = cp.get(session, 'database')
        self.user = cp.get(session, 'user')
        self.password = cp.get(session, 'password')
        self.host = cp.get(session, 'host')
        self.port = int(cp.get(session, 'port'))
        self.connection = None

    def __enter__(self):
        self.connection = pymysql.connect(**{'database': self.database,
                                             'user': self.user,
                                             'password': self.password,
                                             'port': self.port,
                                             'host': self.host})

        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.connection is not None:
            self.connection.close()


class Cursor:
    def __init__(self, connection):
        self.connection = connection
        self._cursor = None

    def __enter__(self):
        self._cursor = self.connection.cursor()
        return self._cursor

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._cursor is not None:
            self._cursor.close()



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: 
# @time: 
# @Software: PyCharm

from abc import ABCMeta
from typing import List

from astartool.error import ParameterTypeError, ParameterValueError
from astartool.number import equals_zero
import numpy as np

from snowland.graphics.core.analytic_geometry_base import Line
from snowland.graphics.core.computational_geometry_2d import Point2D

npa = np.array
npl = np.linalg

__all__ = [
    'Line2D',
    'ConicalSection',
    'Ellipse',
    'Circle',
    'Hyperbola',
    'Parabola',
    'Polynomial',

    'LEFT',
    'ON',
    'RIGHT'
]

LEFT = -1
ON = 0
RIGHT = 1


class Line2D(Line):
    def __init__(self, a=None, b=0, c=0, p1: (Point2D, np.ndarray, list) = None, p2: (Point2D, np.ndarray, list) = None,
                 *args, **kwargs):
        if a is not None and b is not None and c is not None:
            # ax+by+c=0方式定义的线
            self.a, self.b, self.c = a, b, c
        elif p1 is not None and p2 is not None:
            # 两点式定义的直线
            p1, p2 = Point2D(p1), Point2D(p2)
            self.a, self.b, self.c = p2.y - p1.y, p1.x - p2.x, p2.x * p1.y - p1.x * p2.y
        else:
            # TODO: 其他方式确定的直线
            raise ParameterValueError("错误的输入")

    def __distance_point(self, point):
        """
        计算点到直线距离
        :param point: 点
        :return:
        """
        p = Point2D(point)
        return np.fabs(self.a * p.x + self.b * p.y + self.c) / (self.a ** 2 + self.b ** 2) ** 0.5

    def distance_points(self, points):
        """
        计算多点到直线的距离
        :param points: 可以是np.ndarray, 也可以是list(Point2D)
        :return:
        """
        # TODO: 这里可以用矩阵进行运算的，以后补上
        if isinstance(points, list):
            return npa([self.__distance_point(each) for each in points])
        elif isinstance(points, np.ndarray):
            if len(points.shape) == 2:
                return npa([self.__distance_point(each) for each in points])
            else:
                return self.__distance_point(points)
        else:
            raise ValueError('points 参数错误')

    def is_parallel(self, other, eps=1e-8):
        """
        判断直线平行
        """
        if isinstance(other, Line2D):
            return -eps < (self.a - other.a) < eps and -eps < self.b - other.b < eps
        elif isinstance(other, (list, tuple, np.ndarray)):
            return -eps < (self.a - other[0]) < eps and -eps < self.b - other[1] < eps
        else:
            raise ParameterTypeError("输入类型错误")

    def intersection(self, other, eps=1e-8):
        """
        直线交点
        :param other:
        :param eps:
        :return:
        """
        if isinstance(other, Line2D):
            if -eps < (self.a - other.a) < eps and -eps < self.b - other.b < eps:
                return None  # 不相交的两条直线
            else:
                return Point2D([-(self.b * other.c - self.c * other.b) / (other.a * self.b - self.a * other.b), \
                                (self.a * other.c - self.c * other.a) / (other.a * self.b - self.a * other.b)])
        elif isinstance(other, (list, tuple, np.ndarray)):
            assert len(other) == 3
            if -eps < (self.a - other[0]) < eps and -eps < self.b - other[1] < eps:
                return None  # 不相交的两条直线
            return Point2D([-(self.b * other[2] - self.c * other[1]) / (other[0] * self.b - self.a * other[1]), \
                            (self.a * other[2] - self.c * other[0]) / (other[0] * self.b - self.a * other[1])])
        else:
            raise ParameterTypeError("输入类型错误")

    def get(self, x: (float, np.ndarray) = None, y: (float, np.ndarray) = None, eps=1e-10):
        """
        获得对应的取值
        :param x: x值
        :param y: y值
        :param eps: 在此精度下认为是 0
        :return:
        """
        assert (x is None) ^ (y is None), "x和y不全为None"
        if x is None:
            if isinstance(y, (int, float, np.float, np.int)):
                if equals_zero(self.a, eps):
                    return np.inf
                else:
                    return (-self.c - self.b * y) / self.a
            else:
                y = npa(y)
                return (-self.c - self.b * y) / self.a
        else:
            if isinstance(x, (int, float, np.float, np.int)):
                if equals_zero(self.b, eps):
                    return np.inf
                else:
                    return (-self.c - self.a * x) / self.b
            else:
                x = npa(x)
                return (-self.c - self.a * x) / self.b

    def location(self, point: (Point2D, np.ndarray), eps=1e-8):
        point = Point2D(point)
        res = self.a * point.x + self.b * point.y + self.c
        if -eps < res < eps:
            return ON
        elif res > eps:
            return RIGHT
        else:
            return LEFT

    def locations(self, points: (List[Point2D], np.ndarray), eps=1e-8):
        if isinstance(points, list):
            return [self.location(point, eps) for point in points]
        elif isinstance(points, np.ndarray):
            result = np.ones(len(points)) * ON
            res = self.a * points[:, 0] + self.b * points[:, 1] + self.c
            result[res > eps] = RIGHT
            result[res < -eps] = LEFT
            return result
        else:
            raise ParameterTypeError("不支持的points类型")


class ConicalSection(Line, metaclass=ABCMeta):
    """
    圆锥曲线
    """
    pass


class Ellipse(ConicalSection):
    """
    椭圆
    """

    def __init__(self, params, *args, **kwargs):
        # AX2+By2+Cxy+Dx+Ey+F=0
        self.params = params

    @property
    def a(self):
        return self.params[0]

    @property
    def b(self):
        return self.params[1]

    @property
    def c(self):
        return self.params[2]

    @property
    def d(self):
        return self.params[3]

    @property
    def e(self):
        return self.params[4]

    @property
    def f(self):
        return self.params[5]

    @property
    def eccentricity(self):
        e = (self.a / self.b) ** 0.5
        if e > 1:
            return 1 / e
        return e

    @property
    def major_axis(self):
        return max(self.a, self.b) ** 0.5

    @property
    def minor_axis(self):
        return min(self.a, self.b) ** 0.5

    @property
    def focal_length(self):
        return 2 * abs(self.a - self.b) ** 0.5


class Circle(Ellipse):
    def __init__(self, param, *args, **kwargs):
        # x^2+y^2+Dx+Ey+F=0
        if len(param) == 6:
            super(Circle, self).__init__(param)
        elif len(param) == 3:
            p = np.ones(6)
            p[3:] = param
            super(Circle, self).__init__(p)
        else:
            raise ParameterValueError("param参数应为3个或者6个")


class Hyperbola(ConicalSection):
    """
    双曲线
    """

    def __init__(self, params, eps=1e-8, *args, **kwargs):
        # Ax2+2Bxy+Cy2+2Dx+2Ey+F=0
        self.params = params
        if equals_zero(np.linalg.det([[self.a, self.b, self.d],
                                      [self.b, self.c, self.e],
                                      [self.d, self.e, self.f]]), eps):
            raise ParameterValueError("参数矩阵非满秩")

        if self.b * self.b - self.a * self.c <= eps:
            raise ParameterValueError("b^2-ac应该大于0")

    @property
    def a(self):
        return self.params[0]

    @property
    def b(self):
        return self.params[1]

    @property
    def c(self):
        return self.params[2]

    @property
    def d(self):
        return self.params[3]

    @property
    def e(self):
        return self.params[4]

    @property
    def f(self):
        return self.params[5]


class Parabola(ConicalSection):
    """
    抛物线
    """
    def __init__(self, params, eps=1e-8):
        # ax2 +2hxy +by2 +2gx +2fy+c =0
        self.params = params

    @property
    def a(self):
        return self.params[0]

    @property
    def b(self):
        return self.params[2]

    @property
    def h(self):
        return self.params[1]

    @property
    def g(self):
        return self.params[3]

    @property
    def c(self):
        return self.params[5]

    @property
    def f(self):
        return self.params[4]


class Polynomial(Line):
    """
    多项式
    """
    def __init__(self, polynomial=None, coefficient=None, exponent=None):
        """
        coefficient: 系数
        exponent: 指数
        """
        if polynomial is not None:
            if isinstance(polynomial, Polynomial):
                self.polynomial_dict = polynomial.polynomial_dict
            elif isinstance(polynomial, dict):
                self.polynomial_dict = polynomial
            else:
                raise ParameterTypeError("参数类型错误")
        elif coefficient is not None and exponent is not None:
            self.polynomial_dict = dict(zip(exponent, coefficient))
        else:
            self.polynomial_dict = {}

    def diff(self, eps=1e-8):
        """
        导函数
        """
        return Polynomial({k - 1: k * v for k, v in self.polynomial_dict.items() if not (-eps < k < eps)})

    def get(self, x):
        """
        获得x对应的值
        """
        y = np.zeros_like(x)
        for k, v in self.polynomial_dict.items():
            y += v * x ** k
        return y

    def __call__(self, x, *args, **kwargs):
        return self.get(x)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: _ploygon.py
# @time: 2019/8/12 20:48
# @Software: PyCharm

import logging
from typing import List, Union

import numpy as np
from scipy.spatial.ckdtree import cKDTree

from snowland.graphics.core.computational_geometry_2d import Point2D, Polygon, PolygonWithoutHoles, ConvexPolygon
from snowland.graphics.utils import get_intersect_by_two_point
from snowland.graphics.utils import get_rotate_angle_degree, get_angle_rad

npl = np.linalg
npa = np.array
npr = np.random

__all__ = [
    'concave_hull',
    'coss_multi',
    'get_bottom_point_index',
    'graham_scan',
    'in_polygon',
    'on_polygon_edge',
    'sort_polar_angle_by_vector',
    'sort_polar_angle_by_vector_index',
    'sort_polar_angle_cos',
    'sort_polar_angle_cos_index',

]


def __on_polygon_edge(p: Point2D, polygon: (Polygon, list), eps=1e-8):
    p = Point2D(p)
    polygon = Polygon(polygon)
    for point in polygon.p:
        if p == point:  # 使用重载的等于方法
            return True
    edge = np.vstack((polygon.p, polygon.p[0]))
    s1 = np.sign(edge[:-1, 0] - p[0]) * np.sign(edge[1:, 0] - p[0]) < eps
    s2 = np.sign(edge[:-1, 1] - p[1]) * np.sign(edge[1:, 1] - p[1]) < eps
    s = s1 & s2
    if np.any(s):
        # 快速检测
        inds, = np.where(s)
        for each_ind in inds:
            if -eps < get_angle_rad(p.p - edge[each_ind], edge[each_ind + 1] - p.p) < eps:
                return True

    if polygon.holes:
        for hole in polygon.holes:
            hole = PolygonWithoutHoles(hole)
            if on_polygon_edge(p, hole):
                return True
    return False


def on_polygon_edge(point, polygon: (Polygon, np.ndarray)):
    """
    判断点是否在多边形内
    :param point:
    :param polygon:
    :return:
    """
    if isinstance(polygon, (list, np.ndarray)):
        polygon = Polygon(polygon)
    if isinstance(point, Point2D):
        return __on_polygon_edge(point, polygon)
    if isinstance(point, np.ndarray) and len(point.shape) == 1:
        return __on_polygon_edge(Point2D(point), polygon)
    else:
        # 多个点判断， 返回多个值
        if isinstance(point[0], Point2D):
            return npa([__on_polygon_edge(p, polygon) for p in point])
        if isinstance(point[0], np.ndarray) and len(point[0].shape) == 1:
            return npa([__on_polygon_edge(Point2D(p), polygon) for p in point])
        else:
            # 一个点的list/tuple
            return npa([__on_polygon_edge(Point2D(p), polygon) for p in point])


def __in_polygon(p: Point2D, polygon: Polygon):
    """
    判断点是否在多边形内（不考虑边界情况）
    :param point:
    :param polygon:
    :return:
    """
    for point in polygon.p:
        if p == point:  # 使用重载的等于方法
            return True

    for hole in polygon.holes:
        for point in hole:
            if p == point:  # 使用重载的等于方法
                return True
    count = 0
    flag = False
    for i, point in enumerate(polygon.p):  # 注意，这个i取0的时候，i-1值是-1，所以这个是循环到所有的边
        if point[1] <= p[1] < polygon.p[i - 1, 1] or polygon.p[i - 1, 1] <= p[1] < point[1]:
            a, b, c = point[1] - polygon.p[i - 1, 1], polygon.p[i - 1, 0] - point[0], point[0] * polygon.p[i - 1, 1] - \
                      point[1] * polygon.p[i - 1, 0]
            if a > 0:
                if p[0] * a + b * p[1] + c > 0:
                    flag = not flag
                    count += 1
            else:
                if p[0] * a + b * p[1] + c < 0:
                    flag = not flag
                    count += 1
    if polygon.holes:
        for hole in polygon.holes:
            polygon_temp = PolygonWithoutHoles(hole)
            for i, point in enumerate(polygon_temp.p[:, :]):  # 注意，这个i取0的时候，i-1值是-1，所以这个是循环到所有的边
                if (point[1] <= p[1] < polygon_temp.p[i - 1, 1] or polygon_temp.p[i - 1, 1] <= p[1] < point[1]):
                    a, b, c = point[1] - polygon_temp.p[i - 1, 1], polygon_temp.p[i - 1, 0] - point[0], point[0] * \
                              polygon_temp.p[
                                  i - 1, 1] - point[1] * polygon_temp.p[i - 1, 0]
                    if a < 0:
                        if p[0] * a + b * p[1] + c > 0:
                            flag = not flag
                            count += 1
                    else:
                        if p[0] * a + b * p[1] + c < 0:
                            flag = not flag
                            count += 1
    return flag


def in_polygon(point, polygon: (Polygon, np.ndarray)):
    """
    判断点是否在多边形内
    :param point:
    :param polygon:
    :return:
    """
    if isinstance(polygon, (list, np.ndarray)):
        polygon = Polygon(polygon)
    if isinstance(point, Point2D):
        return __in_polygon(point, polygon)
    if isinstance(point, np.ndarray) and len(point.shape) == 1:
        return __in_polygon(Point2D(point), polygon)
    else:
        # 多个点判断， 返回多个值
        if isinstance(point[0], Point2D):
            return npa([__in_polygon(p, polygon) for p in point])
        if isinstance(point[0], np.ndarray) and len(point[0].shape) == 1:
            return npa([__in_polygon(Point2D(p), polygon) for p in point])
        else:
            # 一个点的list/tuple
            return __in_polygon(Point2D(point), polygon)


def get_bottom_point_index(points):
    """
    返回points中纵坐标最小的点的索引，如果有多个纵坐标最小的点则返回其中横坐标最小的那个
    :param points:
    :return:
    """
    n = len(points)
    return min(range(n), key=lambda x: (points[x][-1], points[x][0]))


def sort_polar_angle_by_vector(vectors: np.ndarray, base_vector: np.ndarray):
    return sorted(vectors, key=lambda v: (get_rotate_angle_degree(base_vector, v), -npl.norm(v)))


def sort_polar_angle_by_vector_index(vectors: (np.ndarray, List[Union[np.ndarray, List]]),
                                     base_vector: (np.ndarray, List[Union[float, int]])):
    return sorted(range(len(vectors)),
                  key=lambda v: (get_rotate_angle_degree(base_vector, vectors[v]), -npl.norm(vectors[v])))


def sort_polar_angle_cos_index(points, center_point: (Point2D, np.ndarray), pre_angle=[1, 0]):
    """
    按照与中心点的极角进行排序，使用的是余弦的方法
    :param points: 需要排序的点
    :param center_point: 中心点
    :return:
    """
    if isinstance(center_point, Point2D):
        center_point = center_point.p
    vectors = [npa(each) - npa(center_point) for each in points]
    inds = sort_polar_angle_by_vector_index(vectors, pre_angle)
    return inds


def sort_polar_angle_cos(points, center_point: (Point2D, np.ndarray), pre_angle=[1, 0]):
    """
    按照与中心点的极角进行排序，使用的是余弦的方法
    :param points: 需要排序的点
    :param center_point: 中心点
    :return:
    """
    if isinstance(center_point, Point2D):
        center_point = center_point.p
    vectors = [npa(each) - npa(center_point) for each in points]
    inds = sort_polar_angle_by_vector_index(vectors, pre_angle)
    return [points[ind] for ind in inds]


def coss_multi(v1, v2):
    """
    计算两个向量的叉乘
    :param v1:
    :param v2:
    :return:
    """
    return v1[0] * v2[1] - v1[1] * v2[0]


def graham_scan(points: (np.ndarray, List[np.ndarray])):
    # 使用Graham扫描法计算凸包
    # 网上的代码好多运行效果并不好
    # 算法参见《算法导论》第三版 第605页
    # https://blog.csdn.net/john_bian/article/details/85221039
    if isinstance(points, np.ndarray):
        points = points.tolist()
    bottom_index = get_bottom_point_index(points)
    bottom_point = points.pop(bottom_index)
    sorted_points = sort_polar_angle_cos(points, bottom_point)

    m = len(sorted_points)
    if m <= 2:
        logging.warning("点的数量过少，无法构成凸包")
        return None

    stack = [bottom_point, sorted_points[0], sorted_points[1]]

    for i in range(2, m):
        length = len(stack)
        top = stack[length - 1]
        next_top = stack[length - 2]
        v1 = [sorted_points[i][0] - next_top[0], sorted_points[i][1] - next_top[1]]
        v2 = [top[0] - next_top[0], top[1] - next_top[1]]

        while coss_multi(v1, v2) >= 0:
            stack.pop()
            length = len(stack)
            top = stack[length - 1]
            next_top = stack[length - 2]
            v1 = [sorted_points[i][0] - next_top[0], sorted_points[i][1] - next_top[1]]
            v2 = [top[0] - next_top[0], top[1] - next_top[1]]

        stack.append(sorted_points[i])

    return ConvexPolygon(stack)


def concave_hull(points: np.ndarray, k=15, heigh_number_step=1):
    """计算凹壳(凹包)
    KNN - Moreira and Santos (2007): CONCAVE HULL: A K-NEAREST NEIGHBOURS APPROACH FOR THE COMPUTATION OF THE REGION OCCUPIED BY A SET OF POINTS.
    http://repositorium.sdum.uminho.pt/bitstream/1822/6429/1/ConcaveHull_ACM_MYS.pdf

    """
    dataset_set = {(p[0], p[1]) for p in points}
    dataset = npa(list(dataset_set))
    tree = cKDTree(dataset)
    kk = max(k, 3)  # make sure k>=3
    dist, index = tree.query(dataset, kk + 1)
    if len(dataset) < 3:
        return None  # a minimum of 3 dissimilar points is required
    elif len(dataset) == 3:
        return PolygonWithoutHoles(points)  # for a 3 points points, the polygon is the points itself
    points = dataset
    # kk = min(kk, len(points) - 1)  # make sure that k neighbours can be found
    first_ind = get_bottom_point_index(dataset)
    first_point = points[first_ind]
    used_flag = np.zeros(len(dataset))
    hull = [first_point]  # initialize the hull with the first point 10
    current_ind, current_point = first_ind, first_point
    used_flag[first_ind] = True  # remove the first point 12
    previous_angle, prev_angle = 0, [1, 0]
    step = 1
    while ((current_ind != first_ind) or (step == 1)) and (not np.all(used_flag)):

        if step == 4:
            used_flag[first_ind] = False  # add the first_point again
        avilable_ind = [ind for ind in index[current_ind, 1:] if not used_flag[ind]]

        kNearestPoints = [dataset[ind] for ind in avilable_ind]  # find the nearest neighbours
        cPoints_inds = sort_polar_angle_cos_index(
            kNearestPoints, current_point,
            prev_angle)  # sort the candidates (neighbours) in descending order of right-hand turn
        cPoints = [kNearestPoints[ind] for ind in cPoints_inds]
        curr_ind = None
        for i, p in enumerate(cPoints):
            if avilable_ind[cPoints_inds[i]] == first_ind:
                curr_ind = i
                break
            # select the first candidate that does not intersects any of the polygon edges
            p1, p2 = hull[step - 1], p
            for c1, c2 in zip(hull[:-2], hull[1:-1]):
                if get_intersect_by_two_point(p1, p2, c1, c2) != (None, None):
                    break
            else:
                curr_ind = i
                break
        else:
            return concave_hull(dataset,
                                kk + heigh_number_step)  # since all candidates intersect at least one edge, try again with a higher number of neighbours

        current_point = cPoints[curr_ind]
        hull.append(current_point)  # a valid candidate was found
        prev_angle = hull[step - 1] - hull[step]
        current_ind = avilable_ind[cPoints_inds[curr_ind]]
        used_flag[current_ind] = True
        step += 1
    for i, d in enumerate(dataset):  # check if all the given points are inside the computed polygon 42
        allInside = on_polygon_edge(d, hull) or in_polygon(d, hull)
        if not allInside:
            return concave_hull(dataset,
                                kk + heigh_number_step)  # since at least one point is out of the computed polygon, try again with a higher number of neighbours
    return PolygonWithoutHoles(hull)  # a valid hull was found!


convex_hull = graham_scan

if __name__ == '__main__':
    import numpy as np
    import matplotlib

    from matplotlib import pylab as plt

    r = np.random.random(1000)
    theta = np.random.random(1000) * 2 * np.pi
    ps = np.vstack((r * np.cos(theta), r * np.sin(theta))).T
    hull = convex_hull(ps)
    plt.plot(np.hstack((hull.p[:, 0], hull.p[0, 0])), np.hstack((hull.p[:, 1], hull.p[0, -1])), 'b')
    plt.plot(ps[:, 0], ps[:, 1], 'r*')
    plt.show()



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: qgis_utils.py
# @time: 2022/01/06 20:52
# @Software: PyCharm


from functools import reduce
from typing import List

import numpy as np
from qgis.core import QgsGeometry, QgsPoint, QgsLineString, QgsFeature

from snowland.gis_tool.haversine import length_metres

npa = np.array
npl = np.linalg

__all__ = [
    'points',
    'pointsXY',
    'pointsZM',
    'vector',
    'vector_start',
    'vector_end',
    'point_to_ndarray',
    'start_pointXY',
    'end_pointXY'
]


class GeometryStructure:
    def __init__(self):
        self.geometry_points = np.empty((0, 3))
        self.geometry = None
        self.features = []
        self.road_id = set()
        self.feature_id = []

    @classmethod
    def init_by_features(cls, features: List[QgsFeature]):
        zero_geom_item = GeometryStructure()
        f = features[0]
        zero_geom_item.geometry_points = f.geometry().points()
        zero_geom_item.features.append(f)
        for f in features[1:]:
            points = f.geometry().points()
            zero_geom_item.geometry_points = np.vstack((zero_geom_item.geometry_points, points[1:, :]))
            zero_geom_item.features.append(f)
        zero_geom_item.create_geometry()
        zero_geom_item.create_road_id()
        zero_geom_item.create_id()
        zero_geom_item.create_tile_road_id()
        return zero_geom_item

    def create_geometry(self):
        self.geometry = QgsGeometry(). \
            fromPolyline(QgsLineString(self.geometry_points[:, 0],
                                       self.geometry_points[:, 1],
                                       self.geometry_points[:, 2],
                                       ))

    def create_road_id(self):
        self.road_id = set(f['road_id'] for f in self.features)

    def create_tile_road_id(self):
        self.tile_road_id = [(f['tile_id'], f['road_id']) for f in self.features]

    def create_id(self):
        self.feature_id = set(f['id'] for f in self.features)

    def split_to_geometry(self):
        result = []
        z = self.geometry_points[:, :]
        for feature in self.features:
            ps = feature.geometry().asPolyline()
            y_new = z[:len(ps)]
            z = z[len(ps) - 1:]
            geo_new = QgsGeometry(QgsLineString(y_new[:, 0], y_new[:, 1], y_new[:, 2], np.zeros_like(y_new[:, 0])))
            result.append(geo_new)
        return result

    def reset_geometry(self, geometry: QgsGeometry):
        self.geometry = geometry
        self.geometry_points = geometry.points()


def points(geometry: QgsGeometry):
    vertices = geometry.vertices()
    return npa([(v.x(), v.y(), v.z()) for v in vertices])


def pointsZM(geometry: QgsGeometry):
    vertices = geometry.vertices()
    return npa([(v.x(), v.y(), v.z(), v.m()) for v in vertices])


def pointsXY(geometry: QgsGeometry):
    vertices = geometry.vertices()
    return npa([(v.x(), v.y()) for v in vertices])


def pointsZM(geometry: QgsGeometry):
    vertices = geometry.vertices()
    return npa([(v.x(), v.y(), v.z(), v.m()) for v in vertices])


def vector(geometry: QgsGeometry):
    """
    起点到终点的向量
    """
    start = geometry.vertexAt(0)
    line_nodes = geometry.asPolyline()
    end = line_nodes[-1]
    return npa([end.x() - start.x(), end.y() - start.y()])


def vector_start(geometry: QgsGeometry):
    """
    第一条折线的向量
    """
    start = geometry.vertexAt(0)
    end = geometry.vertexAt(1)
    return npa([end.x() - start.x(), end.y() - start.y()])


def vector_end(geometry: QgsGeometry):
    """
    倒数第一条折线的向量
    """
    line_nodes = geometry.asPolyline()
    start = line_nodes[-2]
    end = line_nodes[-1]
    return npa([end.x() - start.x(), end.y() - start.y()])


def thinning_by_step_method(points, step=2):
    """
    步长法
    :param points:
    :return:
    """
    result = [points[0]]
    delta_length = length_metres(points[:-1], points[1:])
    delta = 0
    for i, d in enumerate(delta_length, start=1):
        if delta + d > step:
            result.append(points[i])
            delta = 0
        else:
            delta += d
    if delta != 0:
        result.append(points[-1])
    return reduce(lambda a, b: np.vstack((a, b)), result)


def douglas_peuker(points, d=5e-7):
    """
    Douglas Peuker
    :param points:
    :return:
    """
    if len(points) < 2:
        return points
    p1 = points[0]
    p2 = points[-1]
    line = npa([p2[1] - p1[1], p1[0] - p2[0], p2[0] * p1[1] - p1[0] * p2[1]])

    dist = np.abs(points[:, 0] * line[0] + points[:, 1] * line[1] + line[2]) / npl.norm(line[:2])
    d_ind = np.argmax(dist)
    if dist[d_ind] < d:
        return npa((p1, p2))
    else:
        return np.vstack((douglas_peuker(points[:d_ind + 1], d), douglas_peuker(points[d_ind:], d)[1:]))


def vertical_distance_limit(points, d=5e-7):
    """
    垂距限值法
    :param points:
    :return:
    """
    p0 = points[0]
    if len(points) < 3:
        return points
    result = [p0]

    for p1, p2 in zip(points[1:-1, :], points[2:, :]):
        line = npa([p2[1] - p0[1], p0[0] - p2[0], p2[0] * p0[1] - p0[0] * p2[1]])
        dist = p1[0] * line[0] + p1[1] * line[1] + line[2]
        print(dist)
        if dist > d:
            result.append(p1)
            p0 = p1
    result.append(points[-1])
    return reduce(lambda a, b: np.vstack((a, b)), result)


def start_pointXY(geometry: QgsGeometry):
    v = geometry.vertexAt(0)
    return npa([v.x(), v.y()])


def end_pointXY(geometry: QgsGeometry):
    v = geometry.asPolyline()[-1]
    return npa([v.x(), v.y()])


def point_to_ndarray(p: QgsPoint):
    return npa([p.x(), p.y()])



# -*- coding: utf-8 -*-

from collections import Counter
import numpy as np

npa = np.array


def histogram(image: np.ndarray):
    """
    像素直方图统计
    :param image:
    :return:
    """
    image: np.ndarray = npa(image)
    return Counter(image.flatten())



# -*- coding: utf-8 -*-

import numpy as np
from astartool.error import ParameterError
from matplotlib import pylab as plt

from snowland.graphics.core.analytic_geometry_2d import Line2D, Polynomial

npa = np.array


def plot_geometry(points, *args, **kwargs):
    points = npa(points)
    plt.plot(points[:, 0], points[:, 1], *args, **kwargs)


def plot_line2d_geometry(line: Line2D, x=None, y=None, *args, **kwargs):
    """
    绘制line_2d
    """
    if x is not None:
        y = line.get(x=x)
        plt.plot(x, y, *args, **kwargs)
    elif y is not None:
        x = line.get(y=y)
        plt.plot(x, y, *args, **kwargs)
    else:
        raise ParameterError("错误的输入数据")


def plot_polynomial(polynomial: Polynomial, interval, points=10, *args, **kwargs):
    """
    绘制多项式
    :param polynomial: 多项式对象
    :param interval: 绘制的区间（左右均为闭区间）
    :param points: 绘制的点数
    :param args: matplotlib参数
    :param kwargs: matplotlib参数
    :return:
    """
    x = np.linspace(interval[0], interval[1], points)
    plt.plot(x, polynomial.get(x), *args, **kwargs)



# -*- coding: utf-8 -*-

import numpy as np
from matplotlib import pylab as plt

from snowland.graphics.utils import rotate_geometry


def plot_line(p1, p2, *args, **kwargs):
    """
    绘制一条线
    """
    plt.plot([p1[0], p2[0]], [p1[1], p2[1]], *args, **kwargs)


def plot_arrow(x, y, *args, **kwargs):
    """
    绘制一个箭头
    """
    ps = np.vstack((x, y)).T
    vec = ps[1:, :] - ps[:-1, :]
    left = rotate_geometry(vec, np.pi / 6) * 0.1
    right = rotate_geometry(vec, -np.pi / 6) * 0.1
    tuple_args = (x, y) + args
    cnt = 0
    for p, l, r in zip(ps[1:], left, right):
        p1, p2 = p - l, p
        tuple_args += ([p1[0], p2[0]], [p1[1], p2[1]]) + args
        p1, p2 = p - r, p
        tuple_args += ([p1[0], p2[0]], [p1[1], p2[1]]) + args
        cnt += 2
    plt.plot(*tuple_args, **kwargs)




# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: qgis_utils.py
# @time: 2022/01/06 20:52
# @Software: PyCharm


__auth__ = 'A.Star'

from typing import Tuple, List

import numpy as np

from snowland.gis_tool import EARTH_RADIUS

npa = np.array


def haversine_centimetres(Olon, Olat, Dlon, Dlat, earth_radius=EARTH_RADIUS):
    """
    通过两经纬度点计算地表实际距离
    """
    d_lat = np.radians(Dlat - Olat)
    d_lon = np.radians(Dlon - Olon)
    a = (np.sin(d_lat / 2.) * np.sin(d_lat / 2.) +
         np.cos(np.radians(Olat)) * np.cos(np.radians(Dlat)) *
         np.sin(d_lon / 2.) * np.sin(d_lon / 2.))
    c = 2. * np.arctan2(np.sqrt(a), np.sqrt(1. - a))
    d = earth_radius * c
    return d * 100000  # 转换到cm为单位


def haversine_metres(Olon, Olat, Dlon, Dlat, earth_radius=EARTH_RADIUS):
    return haversine_centimetres(Olon, Olat, Dlon, Dlat, earth_radius) / 100


def length_centimetres(Olon, Olat, earth_redius=EARTH_RADIUS):
    """
    通过两经纬度点计算地表实际距离

    """
    return np.sum(haversine_centimetres(Olon[:-1], Olat[:-1], Olon[1:], Olat[1:], earth_redius))


def length_metres(Olon, Olat, earth_radius=EARTH_RADIUS):
    """
    通过两经纬度点计算地表实际距离
    """
    return np.sum(haversine_metres(Olon[:-1], Olat[:-1], Olon[1:], Olat[1:], earth_radius))


def haversine_centimetres_by_points(p1, p2, earth_radius=EARTH_RADIUS):
    """
    通过两经纬度点计算地表实际距离
    """
    Olon, Olat = p1[:, 0], p1[:, 1]
    Dlon, Dlat = p2[:, 0], p2[:, 1]
    d_lat = np.radians(Dlat - Olat)
    d_lon = np.radians(Dlon - Olon)
    a = (np.sin(d_lat / 2.) * np.sin(d_lat / 2.) +
         np.cos(np.radians(Olat)) * np.cos(np.radians(Dlat)) *
         np.sin(d_lon / 2.) * np.sin(d_lon / 2.))
    c = 2. * np.arctan2(np.sqrt(a), np.sqrt(1. - a))
    d = earth_radius * c
    return d * 100000  # 转换到cm为单位


def haversine_metres_by_points(p1, p2, earth_radius=EARTH_RADIUS):
    return haversine_centimetres_by_points(p1, p2, earth_radius) / 100


def haversine_metres_by_two_points(p1, p2, earth_radius=EARTH_RADIUS):
    """
    通过两经纬度点计算地表实际距离（单位是米）
    """
    Olon, Olat = p1
    Dlon, Dlat = p2
    d_lat = np.radians(Dlat - Olat)
    d_lon = np.radians(Dlon - Olon)
    a = (np.sin(d_lat / 2.) * np.sin(d_lat / 2.) +
         np.cos(np.radians(Olat)) * np.cos(np.radians(Dlat)) *
         np.sin(d_lon / 2.) * np.sin(d_lon / 2.))
    c = 2. * np.arctan2(np.sqrt(a), np.sqrt(1. - a))
    d = earth_radius * c
    return d * 1000


def haversine_centimetres_by_two_points(p1: (np.ndarray, Tuple, List),
                                        p2: (np.ndarray, Tuple, List),
                                        earth_radius=EARTH_RADIUS):
    """
    通过两经纬度点计算地表实际距离（单位是厘米）
    """
    Olon, Olat = p1
    Dlon, Dlat = p2
    d_lat = np.radians(Dlat - Olat)
    d_lon = np.radians(Dlon - Olon)
    a = (np.sin(d_lat / 2.) * np.sin(d_lat / 2.) +
         np.cos(np.radians(Olat)) * np.cos(np.radians(Dlat)) *
         np.sin(d_lon / 2.) * np.sin(d_lon / 2.))
    c = 2. * np.arctan2(np.sqrt(a), np.sqrt(1. - a))
    d = earth_radius * c
    return d * 100000  # 转换到cm为单位



# -*- coding: utf-8 -*-


from collections import Counter
import numpy as np


def otsu(counter: Counter, colors=256):
    """
    基于统计进行大津法阈值分割
    """
    c_array = np.zeros(colors)
    for k, v in counter.items():
        c_array[k] = v

    p_array = c_array / np.sum(c_array)
    p1 = np.cumsum(p_array)
    range_k = np.arange(colors)
    mk = p_array * range_k
    temp_m = mk / p1
    temp_m[np.isnan(temp_m)] = 0
    m1 = np.cumsum(temp_m)
    mG = m1[-1]
    omega2_b = (mG * p1 - m1) / (p1 * (1 - p1)) * (mG * p1 - m1)
    omega2_b[np.isnan(omega2_b)] = 0
    return np.argmax(omega2_b)




# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file:
# @time:
# @Software: PyCharm

from typing import Iterable

from astartool.error import ParameterError
from scipy.spatial.distance import pdist, cdist, euclidean
import numpy as np
from snowland.graphics.core.computational_geometry_base import Point, LineString, Shape, MultiPoint, Line
from snowland.graphics.utils import get_angle_rad, get_intersect_by_two_point, get_lines, get_arc

npa = np.array
npm = np.mat


__all__ = [
    'Arc2D',
    'Circle',
    'ConvexPolygon',
    'Diamond',
    'Ellipse',
    'LineSegment2D',
    'LineString2D',
    'Point2D',
    'Polygon',
    'PolygonWithoutHoles',
    'Rectangle',
    'Square',
    'Triangle'
]


class Point2D(Point):
    def __init__(self, p=None, x=None, y=None):
        if x is not None and y is not None:
            p = (x, y)
        assert p is not None
        super(Point2D, self).__init__(p)

    @property
    def x(self):
        return self.p[0]

    @x.setter
    def x(self, px):
        self.p[0] = px

    @property
    def y(self):
        return self.p[1]

    @y.setter
    def y(self, py):
        self.p[1] = py


class MultiPoint2D(MultiPoint):
    def __init__(self, points):
        super().__init__(points)
        assert self.p.shape[1] == 2


class MultiPolygon2D(Shape):

    def __init__(self, polygons):
        if isinstance(polygons, MultiPolygon2D):
            self.polygons = polygons.polygons
        elif isinstance(polygons, Iterable):
            self.polygons = [Polygon(p) for p in polygons]

    def area(self):
        return sum(p.area() for p in self.polygons)

    def girth(self):
        return sum(p.girth() for p in self.polygons)


class LineString2D(LineString):
    def __init__(self, X=None):
        super(LineString2D, self).__init__(X=X)
        assert self.X.shape[1] == 2

    def length(self, metric=euclidean, *args, **kwargs):
        return sum(metric(u, v) for u, v in zip(self.X[:-1], self.X[1:]))

    def intersects(self, other):
        if isinstance(other, LineSegment2D):
            lines = get_lines(self.X)
            point_1 = np.hstack((other.X[:-1, :], [[1]])).T
            point_2 = np.hstack((other.X[1:, :], [[1]])).T
            flag1 = (lines @ point_1)
            flag2 = (lines @ point_2)
            if np.all(flag1 * flag2 > 0):
                lines = get_lines(other.X)
                point_1 = np.hstack((self.X[:-1, :], np.ones((len(self.X) - 1, 1)))).T
                point_2 = np.hstack((self.X[1:, :], np.ones((len(self.X) - 1, 1)))).T
                flag1 = (lines @ point_1)
                flag2 = (lines @ point_2)
                if np.all(flag1 * flag2 > 0):
                    return False
            for p1, p2 in zip(self.X[:-1], self.X[1:]):
                if get_intersect_by_two_point(p1, p2, other.X[0], other.X[1]) != (None, None):
                    return True
            return False
        elif isinstance(other, LineString2D):
            for p1, p2 in zip(self.X[:-1], self.X[1:]):
                for p3, p4 in zip(other.X[:-1], other.X[1:]):
                    if get_intersect_by_two_point(p1, p2, p3, p4) != (None, None):
                        return True
            return False
        elif isinstance(other, np.ndarray):
            for p1, p2 in zip(self.X[:-1], self.X[1:]):
                for p3, p4 in zip(other[:-1], other[1:]):
                    if get_intersect_by_two_point(p1, p2, p3, p4) != (None, None):
                        return True
            return False
        else:
            raise ParameterError("type of other error")


class LineSegment2D(LineString2D):
    def __init__(self, X=None, p1=None, p2=None):
        if p1 is not None and p2 is not None:
            if isinstance(p1, Point):
                p1 = p1[:2]
            if isinstance(p2, Point):
                p2 = p2[:2]
            X = np.vstack((p1, p2))
        assert X is not None
        super(LineSegment2D, self).__init__(X)

    def intersection(self, other):
        if isinstance(other, LineSegment2D):
            other = other.X
        return Point2D(get_intersect_by_two_point(self.X[0], self.X[1], other[0], other[1]))

    def intersects(self, other):
        if isinstance(other, LineSegment2D):
            other = other.X
        return Point2D(get_intersect_by_two_point(self.X[0], self.X[1], other[0], other[1])) != (None, None)


class Polygon(Shape):
    def __init__(self, p=None, holes=None):
        self.p = np.zeros((0, 2))
        if isinstance(p, Polygon):
            self.p = p.p
        elif isinstance(p, (list, np.ndarray)):
            if isinstance(p, np.ndarray):
                self.p = p
            else:
                if isinstance(p[0], Point2D):
                    for point in p:
                        self.p = np.vstack((self.p, [point.p]))
                else:
                    for point in p:
                        self.p = np.vstack((self.p, [point]))
            if np.all(self.p[0, :] == self.p[-1, :]):
                self.p = self.p[:-1, :]
        self.holes = []
        if holes:
            for hole in holes:
                if isinstance(hole, np.ndarray):
                    h = hole
                else:
                    h = np.zeros((0, 2))
                    if isinstance(hole[0], Point2D):
                        for point in h:
                            h = np.vstack((h, [point.p]))
                    else:
                        for point in p:
                            h = np.vstack((h, [point]))
                if h[0, :] == h[-1, :]:
                    h = h[:-1, :]
                self.holes.append(h)

    def area(self):
        p = PolygonWithoutHoles(self.p)
        area_p = p.area()
        area_holes = [PolygonWithoutHoles(h).area() for h in self.holes]
        return area_p - np.sum(area_holes)

    def girth(self):
        p = PolygonWithoutHoles(self.p)
        girth_p = p.girth()
        girth_holes = [PolygonWithoutHoles(h).girth() for h in self.holes]
        return girth_p - np.sum(girth_holes)

    def without_holes(self):
        return self.holes is None or not self.holes

    def is_convexhull(self):
        if not self.without_holes():
            return False
        # TODO, check
        raise NotImplementedError


class PolygonWithoutHoles(Polygon):
    def __init__(self, p):
        super(PolygonWithoutHoles, self).__init__(p, holes=None)

    def area(self):
        """
        面积
        :return:
        """
        # TODO 计算面积
        raise NotImplementedError

    def girth(self):
        """
        周长
        :return:
        """
        p_ring = np.vstack((self.p, self.p[:1, :]))
        return LineString(p_ring).length()


class ConvexPolygon(PolygonWithoutHoles):

    def area(self):
        # 面积
        # 参考 https://blog.csdn.net/cymy001/article/details/81366033
        points = np.vstack((self.p, self.p[0]))
        if len(points) < 3:
            raise Exception("error")
        return np.sum(np.fabs(points[:-1, 0] * points[1:, 1] - points[1:, 0] * points[:-1, 1])) / 2


class Triangle(ConvexPolygon):
    """
    三角形
    """
    def __init__(self, p):
        super().__init__(p)
        assert len(self.p) == 3


class Rectangle(ConvexPolygon):
    def __init__(self, p, *args, **kwargs):
        super().__init__(p)
        assert len(self.p) == 4
        vectors = self.p[1:] - self.p[:-1]
        for vi, vj in zip(vectors[:-1], vectors[1:]):
            assert np.isclose(abs(get_angle_rad(vi, vj) * 2), np.pi)


class Diamond(ConvexPolygon):
    def __init__(self, p, *args, **kwargs):
        super(PolygonWithoutHoles, self).__init__(p, holes=None)
        assert len(self.p) == 4
        m, n = self.p.shape
        dist = [pdist(self.p[ind:ind + 2, :]) for ind in range(m - 1)]
        d = cdist(self.p[:1, :], self.p[-1:, :])
        assert np.all(dist == d)


class Square(Rectangle, Diamond):
    """
    正方形
    """

    def __init__(self, p, eps=1e-8, *args, **kwargs):
        super(Square, self).__init__(p, *args, **kwargs)


class Ellipse(Shape):
    def __init__(self, p1, p2, a=0, b=0):
        """
        焦点1, 焦点2, 长轴， 短轴
        """
        self.p1, self.p2 = Point2D(p1), Point2D(p2)
        self.a = a
        self.b = b

    def area(self):
        """
        面积
        :return:
        """
        return self.a * self.b * np.pi

    def girth(self):
        """
        周长
        :return:
        """
        return 4 * self.a * self.b * np.pi

    def is_circle(self, eps=1e-8):
        return -eps <= (self.a - self.b) <= eps


class Circle(Ellipse):
    def __init__(self, centre, r=0):
        super(Circle, self).__init__(centre, centre, r, r)

    @property
    def centre(self):
        return self.p1

    @property
    def r(self):
        return self.a

    @r.setter
    def r(self, radius):
        self.a = self.b = radius


class Arc2D(Line):
    def __init__(self, centre, r, theta_start=-np.pi, theta_end=np.pi, direction=1):
        super().__init__()
        self.centre = centre
        self.r = r
        self.theta_start = theta_start
        self.theta_end = theta_end
        self.direction = direction

        if self.direction > 0:
            assert self.theta_start < self.theta_end
        elif self.direction < 0:
            assert self.theta_start > self.theta_end

    def length(self, metric=euclidean, *args, **kwargs):
        return np.abs(self.theta_end - self.theta_start) * self.r

    def as_linestring2d(self, points_count=30):
        x, y = get_arc(self.centre, self.r, self.theta_start, self.theta_end, points_count)
        points = np.vstack((x, y)).T
        return LineString2D(points)



# -*- coding: utf-8 -*-
from snowland.gis_tool import ELLIPSOID
from snowland.graphics.utils import simple_line, get_rotate_angle_rad, rotate_geometry
from snowland.gis_tool.qgis_tool.distance_area import DistanceArea
import numpy as np

npa = np.array
npl = np.linalg


def move_distance(geo_points, dist, flag, ellipsoid=ELLIPSOID):
    """
    此函数带有端点
    flag: 左侧为-1，右侧为1
    dist: 单位是米
    """
    dist_class = DistanceArea(ellipsoid=ellipsoid)
    if len(geo_points) <= 1:
        return None, None

    firstPoint = 2 * geo_points[0] - geo_points[1]
    lastPoint = 2 * geo_points[-1] - geo_points[-2]
    geo_points = np.vstack((firstPoint, geo_points, lastPoint))

    vectors = geo_points[1:] - geo_points[:-1]
    ps = []
    for i, (vec10, vec12) in enumerate(zip(vectors[:-1], vectors[1:]), start=1):
        vec10_norm = -vec10 / npl.norm(vec10)
        vec12_norm = vec12 / npl.norm(vec10)
        angle = get_rotate_angle_rad(vec10_norm, vec12_norm)
        if flag > 0:
            half_angle = angle / 2
        else:
            half_angle = angle / 2 - np.pi
        vecCenter = rotate_geometry(vec12_norm, half_angle)
        realLen = dist / np.sin(half_angle)

        vecCenter = vecCenter / npl.norm(vecCenter)
        real_point = vecCenter + geo_points[i]
        unitLen = dist_class.distance_metres_by_two_points(geo_points[i], real_point)

        vecCenter = realLen / unitLen * vecCenter if flag > 0 else -realLen / unitLen * vecCenter
        resultVec = vecCenter + geo_points[i]
        ps.append(resultVec)
    ps_array = npa(ps)
    x_array = ps_array[:, 0]
    y_array = ps_array[:, 1]
    return x_array, y_array



# -*- coding: utf-8 -*-


import unittest

import numpy as np
from astartool.project import std_logging
from snowland.graphics.core.analytic_geometry_2d import *

npa = np.array


class TestLine2D(unittest.TestCase):
    @classmethod
    @std_logging()
    def setup_class(cls):
        pass

    @classmethod
    @std_logging()
    def teardown_class(cls):
        print('teardown_class()')

    @std_logging()
    def setup_method(self, method):
        pass

    @std_logging()
    def teardown_method(self, method):
        pass

    def test_line2d(self):
        line2d_1 = Line2D(1, 2, 3)
        line2d_2 = Line2D(p1=(0, 0), p2=(1, 1))
        points = [[0, 0], [0.5, 0.5], [1, 1]]
        res = line2d_2.locations(points)
        for each in res:
            self.assertEqual(ON, each)

        points = npa([[0, 0], [0.5, 0.5], [1, 1]])
        res = line2d_2.locations(points)
        for each in res:
            self.assertEqual(ON, each)

    def test_polynomial(self):
        poly = Polynomial({2: 1, 3: 4})
        new_poly = poly.diff()
        self.assertEqual(new_poly.polynomial_dict, {1: 2, 2: 12})

        poly = Polynomial({4: 1, 3: 4, 0: 6, -1: 7, -2: 4})
        new_poly = poly.diff()
        self.assertEqual(new_poly.polynomial_dict, {3: 4, 2: 12, -2: -7, -3: -8})



from typing import List

import numpy as np
from qgis.core import QgsPoint, QgsLineString, QgsFeature, QgsSpatialIndex

from snowland.gis_tool.qgis_tool import QgsGeometry


class GeometryStructure:
    def __init__(self):
        self.geometry_points = np.empty((0, 4))
        self.geometry = None
        self.features = []
        self.road_id = set()
        self.feature_id = []

    @classmethod
    def init_by_features(cls, features: List[QgsFeature]):
        geom_item = GeometryStructure()
        f = features[0]
        geom_item.geometry_points = f.geometry().pointsZM()
        geom_item.features.append(f)
        for f in features[1:]:
            points = f.geometry().pointsZM()
            geom_item.geometry_points = np.vstack((geom_item.geometry_points, points[1:, :]))
            geom_item.features.append(f)
        geom_item.create_geometry()
        geom_item.create_road_id()
        geom_item.create_id()
        geom_item.create_tile_road_id()
        return geom_item

    def create_geometry(self):
        self.geometry = QgsGeometry(). \
            fromPolyline(QgsLineString(self.geometry_points[:, 0],
                                       self.geometry_points[:, 1],
                                       self.geometry_points[:, 2],
                                       self.geometry_points[:, 3],
                                       ))

    def create_road_id(self):
        self.road_id = set(f['road_id'] for f in self.features)

    def create_tile_road_id(self):
        self.tile_road_id = [(f['tile_id'], f['road_id']) for f in self.features]

    def create_id(self):
        self.feature_id = set(f['id'] for f in self.features)

    def split_to_geometry(self):
        result = []
        z = self.geometry_points[:, :]
        for feature in self.features:
            ps = feature.geometry().asPolyline()
            y_new = z[:len(ps)]
            z = z[len(ps) - 1:]
            geo_new = QgsGeometry(QgsLineString(y_new[:, 0], y_new[:, 1], y_new[:, 2], y_new[:, 3]))
            result.append(geo_new)
        return result

    def reset_geometry(self, geometry: QgsGeometry):
        self.geometry = geometry
        self.geometry_points = geometry.pointsZM()


class Node:
    def __init__(self, data, opposite=None, side=None, right=None):
        self.data = data
        self.opposite, self.side, self.right = opposite, side, right


class TileRoadMap(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for k in self:
            self[k].sort(key=lambda x: -x['lane_no'])

    @staticmethod
    def init_by_features(features):
        waitHandlemap = TileRoadMap()
        for curIndex, feature in enumerate(features):
            key = (feature['tile_id'], feature['road_id'])
            if key in waitHandlemap:
                waitHandlemap[key].append(feature)
            else:
                waitHandlemap[key] = [feature]
        for k in waitHandlemap:
            waitHandlemap[k].sort(key=lambda x: -x['lane_no'])
        return waitHandlemap


class TrafficIntersection:

    def __init__(self):
        self.nodes = {}

    @staticmethod
    def init_by_features(features: List[QgsFeature], buffer_size=1*D2M):
        if not isinstance(features, List):
            features = list(features)
        tile_road_map = TileRoadMap.init_by_features(features)
        zero_index = QgsSpatialIndex()
        last_index = QgsSpatialIndex()
        for v in tile_road_map.values():
            zero_index.addFeature(v[0])
            zero_index.addFeature(v[-1])

        nodes = {k: Node(data=(k, v)) for k, v in tile_road_map.items()}
        for k, v in tile_road_map.items():
            v[0].geometry()



# -*- coding: utf-8 -*-


from typing import Union, List, Iterable

import numpy as np
from astartool.common import BIT_EACH
from astartool.project import time_clock

npa = np.array


def dms_to_nds(dms: Union[np.ndarray, List, float]):
    if isinstance(dms, (np.ndarray, list)):
        nds = ((1 << 32) / 360 * npa(dms)).astype(np.uint64)
    else:
        nds = npa([(1 << 32) / 360 * dms]).astype(np.uint64)
    return nds


def nds_to_morton(nds_lon, nds_lat):
    if isinstance(nds_lon, (np.uint, int)) and isinstance(nds_lat, (np.uint, int)):
        mortonCode = 0
        x, y = nds_lon, nds_lat
        if y < 0:
            y += 0x7FFFFFFF

        y <<= 1

        for i in range(32):
            mortonCode |= x & BIT_EACH[i * 2]
            x <<= 1
            mortonCode |= y & BIT_EACH[i * 2 + 1]
            y <<= 1

        return mortonCode
    if isinstance(nds_lon, list):
        nds_lon = npa(nds_lon)
    if isinstance(nds_lat, list):
        nds_lat = npa(nds_lat)

    assert len(nds_lon) == len(nds_lat)
    bit = 1
    morton_code = np.zeros_like(nds_lat, dtype=np.uint64)
    x, y = nds_lon, nds_lat
    y[y < 0] += 0x7FFFFFFF
    y <<= 1
    for i in range(32):
        morton_code |= x & bit
        x <<= 1
        bit <<= 1
        morton_code |= y & bit
        y <<= 1
        bit <<= 1
    return morton_code


def get_tile_id(lon: (np.ndarray, List[float]), lat: (np.ndarray, List[float]), level=13):
    """
    获得level层的瓦片
    """
    if isinstance(lon, Iterable):
        ndsLon = dms_to_nds(lon)
        ndsLat = dms_to_nds(lat)

        morton = nds_to_morton(ndsLon, ndsLat)
        ntile_id = (((morton >> (2 * (31 - level))) & 0xffffffff) + (1 << (16 + level)))
    else:
        tileids = get_tile_id([lon], [lat], level=level)
        ntile_id = tileids[0]
    return ntile_id


def nds_to_morton2(nds_lon, nds_lat):
    if isinstance(nds_lon, (np.uint, int)) and isinstance(nds_lat, (np.uint, int)):
        mortonCode = 0
        x, y = nds_lon, nds_lat
        if y < 0:
            y += 0x7FFFFFFF

        y <<= 1

        for i in range(32):
            mortonCode |= x & BIT_EACH[i * 2]
            x <<= 1
            mortonCode |= y & BIT_EACH[i * 2 + 1]
            y <<= 1

        return mortonCode
    if isinstance(nds_lon, list):
        nds_lon = npa(nds_lon)
    if isinstance(nds_lat, list):
        nds_lat = npa(nds_lat)

    assert len(nds_lon) == len(nds_lat)
    x, y = nds_lon, nds_lat
    y[y < 0] += 0x7FFFFFFF
    y <<= 1

    x_new = [int(f"{'0'.join(bin(xi)[2:][::-1])}", 2) for xi in x]
    y_new = [int(('0'.join(bin(yi)[2:][::-1])), 2) for yi in y]

    morton_code2 = npa([xi | yi for xi, yi in zip(x_new, y_new)])
    bit = 1
    morton_code = np.zeros_like(nds_lat, dtype=np.uint64)
    xii = np.zeros_like(nds_lat, dtype=np.uint64)
    yii = np.zeros_like(nds_lat, dtype=np.uint64)
    for i in range(32):
        xii |= x & bit
        x <<= 1
        bit <<= 1
        yii |= y & bit
        y <<= 1
        bit <<= 1
    morton_code = xii | yii
    return morton_code2


def get_tile_id2(lon: (np.ndarray, List[float]), lat: (np.ndarray, List[float]), level=13):
    """
    获得level层的瓦片
    """
    if isinstance(lon, Iterable):
        ndsLon = dms_to_nds(lon)
        ndsLat = dms_to_nds(lat)

        morton = nds_to_morton2(ndsLon, ndsLat)
        ntile_id = (((morton >> (2 * (31 - level))) & 0xffffffff) + (1 << (16 + level)))
    else:
        tileids = get_tile_id([lon], [lat], level=level)
        ntile_id = tileids[0]
    return ntile_id


if __name__ == '__main__':
    x = [119.135671785614] * 20
    y = [34.5738355769335] * 20
    s1 = time_clock()
    z = get_tile_id(x, y, 13)
    e1 = time_clock()
    z2 = get_tile_id2(x, y, 13)
    e2 = time_clock()
    print(e1 - s1, e2 - e1)
    for t1, t2 in zip(z, z2):
        assert t1 == 557386867
        assert t2 == 557386867



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: 深圳星河软通科技有限公司 A.Star
# @contact: astar@snowland.ltd
# @site: www.astar.ltd
# @file: graphics .py
# @time: 2020/9/6 23:48
# @Software: PyCharm


import unittest

import numpy as np
from astartool.project import std_logging
from snowland.graphics.core.computational_geometry_2d import *
from snowland.graphics.solid_geometry import in_polygon, on_polygon_edge
from snowland.graphics.utils import interp_points
from snowland.plot_helper.plot_geometry import plot_line2d_geometry
npa = np.array


class TestLine2D(unittest.TestCase):
    @classmethod
    @std_logging()
    def setup_class(cls):
        pass

    @classmethod
    @std_logging()
    def teardown_class(cls):
        print('teardown_class()')

    @std_logging()
    def setup_method(self, method):
        pass

    @std_logging()
    def teardown_method(self, method):
        pass

    def test_line2d(self):
        line2d = LineString2D([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
        segment2d = LineSegment2D(p1=[0, 1], p2=[1, 0])
        self.assertEqual(line2d.length(), 4)
        self.assertTrue(np.isclose(segment2d.length(), np.sqrt(2)))
        self.assertTrue(line2d.is_ring())
        self.assertTrue(line2d.intersects(segment2d))
        segment2d_2 = LineSegment2D(p1=[0.1, 0.1], p2=[0.9, 0.9])
        self.assertFalse(line2d.intersects(segment2d_2))

    def test_in_polygon(self):
        ring = npa([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
        polygon = PolygonWithoutHoles(ring)
        p = [0.5, 0.5]
        self.assertTrue(in_polygon(p, polygon))
        for each in ring:
            self.assertTrue(in_polygon([each], polygon))
        for a, b in zip(ring[:-1], ring[1:]):
            each = (a + b) / 2
            self.assertTrue(on_polygon_edge(each, polygon))
        self.assertFalse(in_polygon([-1, 1], polygon))


class TestComputationalGeometry2D(unittest.TestCase):

    def test_square(self):
        ps = [(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]
        square = Square(ps)
        self.assertTrue(isinstance(square, Diamond))
        self.assertTrue(isinstance(square, Rectangle))
        self.assertEqual(square.area(), 1)
        self.assertEqual(square.girth(), 4)



# -*- coding: utf-8 -*-

from typing import Union, List, Iterable

import numpy as np
from astartool.common import BIT_EACH
from snowland.gis_tool import NDS_180_DEGREES, NDS_360_DEGREES, NDS_90_DEGREES, RULE_MORTON_TO_LONLAT

npl = np.linalg
npa = np.array


def dms_to_nds(dms: Union[np.ndarray, List, float]):
    if isinstance(dms, (np.ndarray, list)):
        nds = ((1 << 32) / 360 * npa(dms)).astype(np.uint64)
    else:
        nds = npa([(1 << 32) / 360 * dms]).astype(np.uint64)
    return nds


def nds_to_morton(nds_lon, nds_lat):
    if isinstance(nds_lon, (np.uint, int)) and isinstance(nds_lat, (np.uint, int)):
        mortonCode = 0
        x, y = nds_lon, nds_lat
        if y < 0:
            y += 0x7FFFFFFF

        y <<= 1

        for i in range(32):
            mortonCode |= x & BIT_EACH[i * 2]
            x <<= 1
            mortonCode |= y & BIT_EACH[i * 2 + 1]
            y <<= 1

        return mortonCode
    if isinstance(nds_lon, list):
        nds_lon = npa(nds_lon)
    if isinstance(nds_lat, list):
        nds_lat = npa(nds_lat)

    assert len(nds_lon) == len(nds_lat)
    bit = 1
    morton_code = np.zeros_like(nds_lat, dtype=np.uint64)
    x, y = nds_lon, nds_lat
    y[y < 0] += 0x7FFFFFFF
    y <<= 1
    for i in range(32):
        morton_code |= x & bit
        x <<= 1
        bit <<= 1
        morton_code |= y & bit
        y <<= 1
        bit <<= 1
    return morton_code


def get_tile_id(lon: (np.ndarray, List[float]), lat: (np.ndarray, List[float]), level=13):
    """
    获得level层的瓦片
    """
    if isinstance(lon, Iterable):
        ndsLon = dms_to_nds(lon)
        ndsLat = dms_to_nds(lat)

        morton = nds_to_morton(ndsLon, ndsLat)
        ntile_id = (((morton >> (2 * (31 - level))) & 0xffffffff) + (1 << (16 + level)))
    else:
        tileids = get_tile_id([lon], [lat], level=level)
        ntile_id = tileids[0]
    return ntile_id


def get_level_of_tile_id(tile_id: int) -> int:
    tId = tile_id >> 16
    level = 0
    while tId:
        tId = tId >> 1
        level += 1
    return level


def get_left_bottom_of_tile(tile_id, level=None):
    if level is None:
        level = get_level_of_tile_id(tile_id)
    return get_left_bottom_of_tile_with_level(tile_id, level)


def get_index_of_tile_id_with_level(tile_id: int, level: int):
    offset = 1 << (16 + level)
    return tile_id - offset


def normalize_coord(result):
    """纠偏"""
    # if x > 180 degrees, then subtract 360 degrees
    if result[0] > NDS_180_DEGREES:
        result[0] -= NDS_360_DEGREES + 1  # add 1 because 0 must be counted as well
    elif result[0] < -NDS_180_DEGREES:  # if x < 180 , x += 360
        result[0] += NDS_360_DEGREES + 1  # add 1 because 0 must be counted as well

    # if y > 90 degrees, then subtract 180 degrees
    if result[1] > NDS_90_DEGREES:
        result[1] -= NDS_180_DEGREES + 1  # add 1 because 0 must be counted as well
    elif result[1] < -NDS_90_DEGREES:  # if y < 90, y += 180
        result[1] += NDS_180_DEGREES + 1  # add 1 because 0 must be counted as well
    return result


def morton_code_to_coord(morton_code_param):
    coord = np.zeros(2, dtype=np.uint32)
    morton_code = morton_code_param
    bit = 1
    for i in range(32):
        coord[0] = coord[0] | (morton_code & bit)
        morton_code = morton_code >> 1
        coord[1] = coord[1] | (morton_code & bit)
        bit = bit << 1
    return coord


def get_left_bottom_of_tile_with_level(tile_id: int, level: int = None):
    if level is None:
        level = get_level_of_tile_id(tile_id)
    indexOfTile = get_index_of_tile_id_with_level(tile_id, level)
    mortonCode = indexOfTile << (2 * (31 - level))
    coord = morton_code_to_coord(mortonCode)
    coord = normalize_coord(coord)
    coord += 1
    return coord


def get_lon_lat_of_tile(tile_id: int, level: int = None):
    if level is None:
        level = get_level_of_tile_id(tile_id)
    coord = get_left_bottom_of_tile_with_level(tile_id, level)
    result = coord * RULE_MORTON_TO_LONLAT
    return result


def get_tile_bounding_box(tile_id: int, level=13):
    """
    获得tile所在的矩形框
    """
    if level is None:
        level = get_level_of_tile_id(tile_id)
    p = get_left_bottom_of_tile(tile_id, level)
    length = 180 / (1 << level)
    return p, length, length


def get_a_set_of_tiles(start_point, end_point, level=13):
    """
    获取一系列的tile快
    :param start_point:
    :param end_point:
    :param level:
    :return:
    """
    points: np.ndarray = np.vstack((start_point, end_point))
    start_point, end_point = points.min(axis=0), points.max(axis=0)
    start, _, _ = get_left_bottom_of_tile(get_tile_id(*start_point, level=level), level=level)
    end, len1, len2 = get_left_bottom_of_tile(get_tile_id(*end_point, level=level), level=level)
    end[0] += len1
    end[1] += len2
    xi = np.arange(start[0], end[0], len1)
    yi = np.arange(start[1], end[1], len2)
    XI, YI = np.meshgrid(xi, yi)
    points = np.concatenate((np.expand_dims(XI, axis=0), np.expand_dims(YI, axis=0)), axis=0)
    points_all = np.expand_dims(points, axis=0)
    points0 = points_all[:, :, :-1, :-1]
    points1 = points_all[:, :, 1:, :-1]
    points2 = points_all[:, :, 1:, 1:]
    points3 = points_all[:, :, :-1, 1:]
    blocks = np.concatenate((points0, points1, points2, points3, points0), axis=0)
    return blocks
